@@ -1,9 +1,10 @@
 from raachem.file_class.xyz import XyzFile
 from raachem.util.constants import elements
 from raachem.util.gen_purp import is_str_float
-import functools
+import functools, itertools, math
 
 class LogFile:
+	calc_types = ["TS","Red","IRC","Opt","SP"]
 	def __init__(self,file_content):
 		self.list = file_content
 		self.lenght = len(self.list)
@@ -18,8 +19,7 @@ class LogFile:
 		self.multi_dash_idxs =[]
 		self.scf_done = []
 		####.thermal = ["ZPC","TCE","TCH","TCG","SZPE","STE","STH","STG"]
-		self.thermal = [None , None, None, None, None , None, None, None]
-		self.start_displ_block = None
+		self.thermal = [None, None, None, None, None , None, None, None]
 		self.oc_orb_energies = []
 		self.uno_orb_energies = []
 		self.hash_line_idxs = []
@@ -28,6 +28,15 @@ class LogFile:
 		self.irc_points = []
 		self.scan_points = []
 		self.opt_points = []
+		self.force_const_mat = []
+		self.distance_matrix = []
+		self.s_squared = []
+		self.muliken_spin_densities_idxs = []
+		self.muliken_charge_idxs = []
+		self.chelpg_charge_idxs = []
+		self.pop_analysis_idxs = []
+		self.npa_start_idxs = []
+		self.npa_end_idxs = []
 		for i,a in enumerate(a.strip() for a in self.list):
 			# i = index
 			# a = line.strip()
@@ -48,27 +57,40 @@ class LogFile:
 			elif a[0] == "A":
 				text_a = "Alpha  occ. eigenvalues --"
 				text_b = "Alpha virt. eigenvalues --"
+				text_c = "Atom  No          Natural Electron Configuration"
 				if a.startswith(text_a):
-					text = a.replace(text_a,"").split();                     self.oc_orb_energies.extend(text); continue
+					text = a.replace(text_a,"").replace("-"," -").split();   self.oc_orb_energies.extend(text); continue
 				elif a.startswith(text_b):
 					text = a.replace(text_b,"").split();                    self.uno_orb_energies.extend(text); continue
+				elif a.split() == text_c.split():
+					pass;                                                          self.npa_end_idxs.append(i); continue
+
 			elif a[0] == "C":
 				b = a.split(); c = len(b)
 				if all((a.startswith("Charge"),self.charge_mult is None, c == 6)):
 					pattern = ("Charge", "=", "Multiplicity", "=")
 					if all(x == b[n] for x,n in zip(pattern,(0,1,3,4))):
 						self.input_geom_idx = i;                                    self.charge_mult = b[2::3]; continue
+			elif a[0] == "D":
+				if a.startswith("Distance matrix (angstroms):"):   				self.distance_matrix.append(i); continue
 			elif a[0] == "E":
 				if a.startswith("Error"):                                                self.errors.append(i); continue
+				elif a.startswith("ESP charges:"):                           self.chelpg_charge_idxs.append(i);continue
 			elif a[0] == "F":
-				condition_a = a.startswith("Frc consts  -- ") and self.start_displ_block is None
-				if condition_a:                                                 self.start_displ_block = i - 4; continue
+				if a.startswith("Full mass-weighted force constant matrix:"):   self.force_const_mat.append(i); continue
 			elif a[0] == "I":
 				if a == "Input orientation:":                                self.start_xyz_idxs.append(i + 5); continue
 			elif a[0] == "L":
 				if a.startswith("Link1:"):                                      self.linked_job_idxs.append(i); continue
 			elif a[0] == "N":
 				if a.startswith("Normal termination of Gaussian"):               self.norm_term_idxs.append(i); continue
+			elif a[0] == "M":
+				if a.startswith("Mulliken charges and spin densities:"):
+					pass;                                           self.muliken_spin_densities_idxs.append(i); continue
+				elif a.startswith("Mulliken charges:"):
+					pass;                                                    self.muliken_charge_idxs.append(i);continue
+				elif a.startswith("Molecular Orbital Coefficients:"):
+					pass;                                                      self.pop_analysis_idxs.append(i);continue
 			elif a[0] == "P":
 				b = a.split(); c = len(b)
 				if c != 6 or any(x != b[n] for x,n in zip(["Point","Number:","Path","Number:"],[0,1,3,4])):     continue
@@ -78,6 +100,7 @@ class LogFile:
 				b = a.split(); c = len(b)
 				if a == "Standard orientation:":                             self.start_xyz_idxs.append(i + 5); continue
 				elif a.startswith("SCF Done:") and c > 5:                       self.scf_done.append([i,b[4]]); continue
+				elif a.startswith("S**2 before annihil"):self.s_squared.append([i,b[3].replace(",",""),b[-1]]); continue
 				elif a.startswith("Sum of electronic and zero-point Energies="):       self.thermal[4] = b[-1]; continue
 				elif a.startswith("Sum of electronic and thermal Energies="):          self.thermal[5] = b[-1]; continue
 				elif a.startswith("Sum of electronic and thermal Enthalpies="):        self.thermal[6] = b[-1]; continue
@@ -96,16 +119,29 @@ class LogFile:
 				b = a.split()
 				if a.startswith("Zero-point correction="):                             self.thermal[0] = b[-2]; continue
 			elif a[0] == "#":                                                    self.hash_line_idxs.append(i); continue
+			elif a[0] == "*":
+				if a.replace("*","").startswith("Gaussian NBO Version 3.1"):
+					pass;                                                         self.npa_start_idxs.append(i);continue
 		#--------------------------------------------POST PROCESSING----------------------------------------------------
 		x = None if self.start_xyz_idxs is None else [min(a for a in self.multi_dash_idxs if a > b) for b in self.start_xyz_idxs]
 		self.end_xyz_idxs = x
 		self.scan_end = [min(a for a in self.multi_dash_idxs if a > b[0]) for b in self.scan_points]
-		x = None if self.start_displ_block is None else min(a for a in self.empty_line_idxs if a > self.start_displ_block)
-		self.end_displ_block = x
-		self.displ_block = None if self.end_displ_block is None else self.list[self.start_displ_block:self.end_displ_block]
-		x = None if self.hash_line_idxs is None else min(a for a in self.multi_dash_idxs if a > self.hash_line_idxs[0])
-		x = None if self.hash_line_idxs is None else "".join([a.lstrip() for a in self.list[self.hash_line_idxs[0]:x]])
-		self.raw_route = x
+		try:
+			x = [self.list[b:min(a for a in self.empty_line_idxs if a > b)] for b in self.force_const_mat]
+			self.displ_block = x
+		except Exception as e:
+			print("Error while finding vibrational frequencies of log file")
+			print(e)
+			print(self.name)
+			self.displ_block = []
+		try:
+			x = None if self.hash_line_idxs is None else min(a for a in self.multi_dash_idxs if a > self.hash_line_idxs[0])
+			x = None if self.hash_line_idxs is None else "".join([a.lstrip() for a in self.list[self.hash_line_idxs[0]:x]])
+			self.raw_route = " ".join(x.split())
+		except IndexError as e:
+			print("Error while finding route section of log file")
+			print(e)
+			print(self.name)
 		try:
 			self.oc_orb_energies = [float(b) for b in self.oc_orb_energies]
 			self.uno_orb_energies = [float(a) for a in self.uno_orb_energies]
@@ -139,10 +175,8 @@ class LogFile:
 				self.init_errors.append("Lumo is lower than homo?")
 		if self.init_errors:
 			for a in self.init_errors: print(a)
-			print("Errors above were found on file\n {}".format(self.name))
-		#self.loghelp()
-		#print(self.raw_route)
-		#print(self.first_xyz_obj())
+			print("Errors above were found on file\n{}".format(self.name))
+		#self._xyz_from_dist_matrix()
 	@functools.lru_cache(maxsize=1)
 	def loghelp(self):
 		for a in vars(self):
@@ -156,10 +190,10 @@ class LogFile:
 	def last_cord_block(self):
 		if not all([self.xyz_cord_block, self.end_xyz_idxs]):
 			if self.resumes:
-				print("WARNING: Coordinates will be drawn from the last job abstract:")
+				print("WARNING: Coordinates will be taken from the last job abstract:")
 				print("lines {} - {} of file:".format(self.start_resume_idxs[-1],self.end_resume_idxs[-1]))
 				print("{}".format(self.name))
-				return LogAbstract(self.resumes[0]).xyz_object().cord_block()
+				return self.last_log_abstract.xyz_object().cord_block()
 			else: return None
 		else:
 			return self.xyz_cord_block(self.start_xyz_idxs[-1],self.end_xyz_idxs[-1])
@@ -186,36 +220,37 @@ class LogFile:
 			return len(self.last_cord_block())
 		elif self.first_cord_block():
 			return len(self.first_cord_block())
+
 	def any_xyz_obj(self,a_idx,b_idx,title=" ",name=False):
 		if name == False: name = self.name
 		return XyzFile([name, self.n_atoms, title, *(" ".join(l) for l in self.xyz_cord_block(a_idx,b_idx))])
 	@functools.lru_cache(maxsize=1)
 	def last_xyz_obj(self):
-		return XyzFile([self.name,self.n_atoms," ",*(" ".join(l) for l in self.last_cord_block())])
+		if self.last_cord_block():
+			return XyzFile([self.name,self.n_atoms," ",*(" ".join(l) for l in self.last_cord_block())])
+		else:
+			return None
+
 	@functools.lru_cache(maxsize=1)
 	def first_xyz_obj(self):
 		return XyzFile([self.name,self.n_atoms," ",*(" ".join(l) for l in self.first_cord_block())])
+
 	@functools.lru_cache(maxsize=1)
-	def frequencies(self):
-		if self.displ_block is None: return False
-		else: return [j for a in [i.split()[2:] for i in self.displ_block[2::self.n_atoms+7]] for j in a]
-	@functools.lru_cache(maxsize=1)
-	def displ_for_freq_idx(self,freq_idx):
-		if self.displ_block is None: return []
-		displ = []
-		for num in range(self.n_atoms):
-			displ.append([j for a in [i.split()[2:] for i in self.displ_block[7+num::self.n_atoms+7]] for j in a])
-		displ_for_freq_str = [a[freq_idx*3:freq_idx*3+3] for a in displ]
-		displ_for_freq_float = [[float(i) for i in b] for b in displ_for_freq_str]
-		return displ_for_freq_float
+	def low_e_xyz_obj(self):
+		if self.calc_type == "SP": return None
+		else:
+			xyzs = {"TS":self.opt,"Red":self.scan_geoms,"IRC":self.irc,"Opt":self.opt}[self.calc_type]()
+			if len(xyzs) == 0: return None
+			else: return sorted(xyzs,key= lambda x: float(x.title()) if is_str_float(x.title()) else 1)[0]
+
 	@functools.lru_cache(maxsize=1)
 	def _calc_type(self):
 		if self.raw_route:
 			r_sect = [self.raw_route]
 			for x in [None, "/", "(", ")", ",", "=", "%", ":"]:
-				r_sect = [a for b in [i.split(x) for i in r_sect] for a in b if len(a) > 1]
+				r_sect = [a for a in itertools.chain(*[i.split(x) for i in r_sect]) if len(a) > 1]
 			r_sect = [a.lower() for a in r_sect]
-			if "ts" in r_sect: return "TS"
+			if any(a in r_sect for a in ("ts", "qst2","qst3")): return "TS"
 			elif any(True for a in r_sect if a in ("modredundant", "readoptimize", "readfreeze")): return "Red"
 			elif "irc" in r_sect: return "IRC"
 			elif "opt" in r_sect: return "Opt"
@@ -226,9 +261,22 @@ class LogFile:
 		return any(True if "Normal termination of Gaussian" in l else False for l in self.list[-5:])
 	def _error_msg(self):
 		error_idxs = [a for a in self.errors if a + 5 > self.lenght]
+		for n in [-4,-3,-2,-1]:
+			if self.list[n].strip().startswith("galloc:  could not allocate memory."):
+				error_idxs.append(n)
 		if error_idxs: return " | ".join([self.list[n] for n in error_idxs])
 		else: return "No data"
 	@functools.lru_cache(maxsize=1)
+	def needs_ref(self):
+		if self.calc_type == "Opt" and self.last_freq:
+			if self.last_freq.n_ifreq() == "0": return "No"
+			else: return "Yes"
+		elif self.calc_type == "TS" and self.last_freq:
+			if self.last_freq.n_ifreq() == "1": return "No"
+			else: return "Yes"
+		else: return "-"
+
+	@functools.lru_cache(maxsize=1)
 	def irc(self):
 		if not all([self.start_xyz_idxs,self.end_xyz_idxs,self.irc_points,self.scf_done]): return []
 		points = self.irc_points
@@ -251,11 +299,15 @@ class LogFile:
 	def scan_geoms(self):
 		if not all([self.start_xyz_idxs, self.end_xyz_idxs, self.scan_points, self.scf_done]): return []
 		geoms = []
-		points = self.scan_points
-		start_idx = [min(i for i in self.start_xyz_idxs if i > b[0]) for b in points]
+		all_points = self.scan_points
+		points = [a for a in all_points if a[0] < self.start_xyz_idxs[-1] and a[0] < self.end_xyz_idxs[-1]]
+		points_removed = len(all_points) - len(points)
+		if points_removed != 0:
+			print(f"WARNING: {points_removed} Scan  points have been removed due to inconsistent number o geometries found")
+		start_idx = [min(i for i in self.start_xyz_idxs if i > b[0])  for b in points]
 		end_idx = [min(i for i in self.end_xyz_idxs if i > b[0]) for b in points]
 		scf_idx = [max(i for i in self.scf_done if i[0] < b[0]) for b in points]
-		for i,(a,b,c,d) in enumerate(zip(start_idx,end_idx,scf_idx,self.scan_points)):
+		for i,(a,b,c,d) in enumerate(zip(start_idx,end_idx,scf_idx,points)):
 			name = self.name.replace(".log","_" + str(i+1)+".xyz")
 			if d[1] == "Optimized": print("Optimized geometry found at line {}!".format(d[0]))
 			elif d[1] == "Non-Optimized": print("Non-Optimized1 geometry found at line {}!".format(d[0]))
@@ -263,14 +315,70 @@ class LogFile:
 		if len(geoms) == 0:
 			print("No Optimized geometries found for {} file".format(self.name()))
 		return geoms
-
+	@functools.lru_cache(maxsize=1)
+	def _last_freq(self):
+		return LogFreq(self.displ_block[-1]) if self.displ_block else False
+	@functools.lru_cache(maxsize=1)
+	def _last_log_abstract(self):
+		return LogAbstract(self.resumes[-1]) if self.resumes is not None else False
+	def _xyz_from_dist_matrix(self):
+		end_idx = lambda x: next(i for i,a in enumerate(self.list[x+1:],start=x+1) if not a.split()[0].isdigit())
+		return [DistMatrix(self.list[a+1:end_idx(a)]) for a in self.distance_matrix]
+	def _last_muliken_spin_density(self):
+		if self.muliken_spin_densities_idxs:
+			end_idx = lambda x: next(i for i, a in enumerate(self.list[x + 1:], start=x + 1) if not a.split()[0].isdigit())
+			return "\n".join(self.list[self.muliken_spin_densities_idxs[-1]:end_idx(self.muliken_spin_densities_idxs[-1]+1)])
+		else:
+			return "-"
+	def _last_internal_coord(self):
+		if self.scan_points:
+			end_idx = lambda x: next(i for i, a in enumerate(self.list[x + 1:], start=x + 1) if not a.strip().startswith("!"))
+			return "\n".join(self.list[self.scan_points[-1][0]-1:end_idx(self.scan_points[-1][0]+5)+1])
+		else:
+			return "-"
+	def _last_muliken_charges(self):
+		if self.muliken_charge_idxs:
+			end_idx = lambda x: next(i for i, a in enumerate(self.list[x + 1:], start=x + 1) if not a.split()[0].isdigit())
+			return "\n".join(self.list[self.muliken_charge_idxs[-1]:end_idx(self.muliken_charge_idxs[-1]+1)])
+		else:
+			return "-"
+	def _last_chelpg_charges(self):
+		if self.chelpg_charge_idxs:
+			end_idx = lambda x: next(i for i, a in enumerate(self.list[x + 1:], start=x + 1) if not a.split()[0].isdigit())
+			return "\n".join(self.list[self.chelpg_charge_idxs[-1]:end_idx(self.chelpg_charge_idxs[-1] + 1)])
+		else:
+			return "-"
+	def _pop_analysis(self):
+		if self.pop_analysis_idxs and self.muliken_charge_idxs:
+			return "\n".join(self.list[self.pop_analysis_idxs[-1]:self.muliken_charge_idxs[-1]])
+		else:
+			return "-"
+	def _npa_analysis(self):
+		if self.npa_start_idxs and self.npa_end_idxs:
+			if len(self.npa_start_idxs) > 1:
+				end_idx = lambda x: next(i for i, a in enumerate(self.list[x + 1:], start=x + 1) if a.strip() == "")
+				return "\n".join(self.list[self.npa_start_idxs[-2]:end_idx(self.npa_end_idxs[-1])])
+			else:
+				return "-"
+		else:
+			return "-"
 	n_atoms = property(_n_atoms)
 	normal_termin = property(_normal_termin)
 	calc_type = property(_calc_type)
 	error_msg = property(_error_msg)
+	last_log_abstract = property(_last_log_abstract)
+	last_freq = property(_last_freq)
+	xyz_from_dist_matrix = property(_xyz_from_dist_matrix)
+	last_muliken_spin_density = property(_last_muliken_spin_density)
+	last_internal_coord = property(_last_internal_coord)
+	last_muliken_charges = property(_last_muliken_charges)
+	last_chelpg_charges = property(_last_chelpg_charges)
+	pop_analysis = property(_pop_analysis)
+	npa_analysis = property(_npa_analysis)
 
 class LogAbstract:
 	def __init__(self,content):
+		assert type(content) is list
 		self.list = content
 		self.version = None
 		self.dipole = None
@@ -305,3 +413,86 @@ class LogAbstract:
 	def xyz_object(self):
 		return self.read_strucure()[1]
 
+class LogFreq:
+	def __init__(self, content):
+		assert type(content) is list
+		self.list = content
+		self.rows = []
+		for i,a in enumerate(self.list):
+			if a.lstrip().startswith("Frequencies --"):
+				try:
+					assert self.list[i + 1].lstrip().startswith("Red. masses --")
+					assert self.list[i + 2].lstrip().startswith("Frc consts  --")
+					assert self.list[i + 3].lstrip().startswith("IR Inten    --")
+					self.rows.append(i-2)
+				except AssertionError:
+					continue
+		if not self.rows:
+			self.n_atoms = 1
+			self.block = []
+		else:
+			self.n_atoms = len(self.list) - self.rows[-1] - 7
+			self.block = self.list[self.rows[0]:]
+	def __str__(self):
+		return "\n".join(self.list)
+	@functools.lru_cache(maxsize=1)
+	def frequencies(self):
+		return list(itertools.chain(*[i.split()[2:] for i in self.block[2::self.n_atoms+7]]))
+	@functools.lru_cache(maxsize=1)
+	def ir_intensities(self):
+		return list(itertools.chain(*[i.split()[3:] for i in self.block[5::self.n_atoms+7]]))
+	@functools.lru_cache(maxsize=1)
+	def displ_for_freq_idx(self,freq_idx):
+		displ = []
+		for num in range(self.n_atoms):
+			displ.append(list(itertools.chain(*[i.split()[2:] for i in self.block[7+num::self.n_atoms+7]])))
+		displ_for_freq_str = [a[freq_idx*3:freq_idx*3+3] for a in displ]
+		displ_for_freq_float = [[float(i) for i in b] for b in displ_for_freq_str]
+		return displ_for_freq_float
+	@functools.lru_cache(maxsize=1)
+	def n_ifreq(self):
+		return str(len([b for b in self.frequencies() if float(b) < 0])) if self.frequencies() else "No data"
+	def ir_spectra(self,threshold = 20):
+		pairs = []
+		for a,b in zip(self.frequencies(), self.ir_intensities()):
+			if is_str_float(a) and is_str_float(b):
+				pairs.append([float(a),float(b)])
+		for a,b in zip(sorted(pairs,key=lambda x: x[1], reverse=True), range(threshold)):
+			print("{:>10.1f}{:>10.1f}".format(float(a[0]),float(a[1])))
+		print("---------")
+class DistMatrix:
+	def __init__(self,text):
+		self.element = {}
+		for a in [b.split() for b in text]:
+			idx = "".join(a[0:2])
+			if len(a) > 2 and a[1] in elements:
+				if idx in self.element:
+					self.element[idx].extend([float(c) for c in a[2:]])
+					continue
+				else:
+					self.element[idx] = [float(c) for c in a[2:]]
+		for a in self.element:
+			print(a,self.element[a])
+		self.dist_matrix = sorted(self.element.values(),key=lambda x: len(x))
+		self.elem_vector = sorted(self.element.keys(), key=lambda x: len(self.element[x]))
+		self.xyz_ent_a = []
+		self.xyz_ent_b = []
+		for i,a in enumerate(self.dist_matrix):
+			if i == 0:
+				self.xyz_ent_a.append([0, 0, 0])
+				self.xyz_ent_b.append([0, 0, 0])
+			if i == 1:
+				self.xyz_ent_a.append([a[0], 0, 0])
+				self.xyz_ent_b.append([a[0], 0, 0])
+			if i == 2:
+				x = (self.dist_matrix[i-1][0]**2+a[0]**2-a[1]**2)/(2*self.dist_matrix[i-1][0])
+				y = math.sqrt(a[1]**2-x**2)
+				self.xyz_ent_a.append([x, y, 0])
+				self.xyz_ent_b.append([x, y, 0])
+			if i > 2:
+				x = (self.dist_matrix[i-1][0]**2+a[0]**2-a[1]**2)/(2*self.dist_matrix[i-1][0])
+				y = math.sqrt(a[1]**2-x**2)
+				#z =
+				self.xyz_ent_a.append([x, y, 0])
+				self.xyz_ent_b.append([x, y, 0])
+
